/* Functions for the game (No UI)
 * $Id: game.c,v 1.22 2000/04/27 23:25:56 marijn Exp $

 * An air traffic simulation game. 
 * Copyright (C) 2000  Marijn Vriens <marijn@sanity.dhs.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>

#include "defaults.h"
#include "datastructs.h"
#include "tests.h"
#include "utils.h"
#include "game.h"
#include "lists.h"
#include "interface.h"

gboolean
processPlane(airPlane *, airSetting *);

gboolean
processObject(airObj *, airSetting *);

gint
planeTurn(gint, gint, gint);

gint
planeChangeHeight(gint, gint);

void
planeScore(airPlane *, airObj *);

void
planeNearMiss(GList *);

airPlane *
planeSetup(airPlane *);

airPlane *
planeReset(airPlane *);

airPlane *
planeClearCmd(airPlane *);

GList*
planeProximity(GList *, gint);

gint
planeLeave(airPlane *plane, airObj **exitobj);



void /* the game round, gets called when a round is started */
gameRound()
{
	GList *curNode   = NULL;
	GList *nextNode  = NULL;
	GList *firstNode = NULL;

	GList *missPlanes  = NULL;
	GList *crashPlanes = NULL;

	airPlane *removeData = NULL;
	guint plnNr, i;

	/* check that globals are sane and exist */
	g_assert(airSet != NULL);
	g_assert(airObjs != NULL);
	g_assert(airPlanes != NULL);
        /* we shouldn't be here if there are no more planes to fly */
	
	/* rewind global lists */
  	airObjs   = g_list_first(airObjs);
	airPlanes = g_list_first(airPlanes);
	if(airPlanesFinish != NULL) {
		airPlanesFinish = g_list_first(airPlanesFinish);
	}
	
	/* act on the obects in the game */
	curNode = firstNode = airObjs;
   	do { 
  		processObject(curNode->data, airSet); 
	} while((curNode = g_list_next(curNode)) != NULL); 
	airObjs = firstNode;
	curNode = firstNode = NULL;

        /* act on the planes in the game */
	curNode = airPlanes;
	plnNr = g_list_length(curNode);
	for(i=0;TRUE;i++) {
		nextNode = g_list_next(curNode);
		if(processPlane(curNode->data, airSet) == FALSE) {
			/* curNode points to the node needing removal */
			removeData = curNode->data;
			curNode = g_list_remove(curNode, removeData);
			airPlanesFinish = g_list_prepend(airPlanesFinish, removeData);			
		}
		if(nextNode != NULL) {
			curNode = nextNode;
		} else {
			break;
		}

	}	
	g_list_first(curNode); /* rewind list */
	airPlanes = curNode;   /* commit changed list to global */
	curNode = nextNode = NULL;
	
	if(airPlanes != NULL) {
		crashPlanes = planeProximity(airPlanes, airSet->crashDist);
		missPlanes  = planeProximity(airPlanes, airSet->missDist);
		if(crashPlanes != NULL) {
			g_print("PLANES CRASHED!\n");
			build_crash_board();
			gameEnd(TRUE);
//		g_assert_not_reached();
		}
		if(missPlanes != NULL) {
			planeNearMiss(missPlanes);
		}
	}
}

void
gameStart() {
	/* reset administrative values */
	airSet->stat.roundNr = 0;
	airSet->stat.progress= 0;

	/* clean up after old games */

	/* reset scores */
	airSet->score.points  = 0;
	airSet->score.correct = 0;
	airSet->score.nearmiss= 0;
	
	gameEnd(FALSE);
	
	/* let the games begin */
	airObjs   = genObjList( airSet );
	airPlanes = genPlaneList( airSet, airObjs );
	loadFlightList( GTK_CLIST(airUi->flist), airPlanes );
	game_timer_start();
}

void /* the game has ended, do your thing */
gameEnd(gboolean showScore) {
	if(showScore) 
		build_result_board(&(airSet->score));
	
	game_timer_stop();
	airPlanesFinish = freePlaneList( airPlanesFinish );
	if(airPlanes != NULL) { /* if there was a list with airplanes */
		airPlanes = freePlaneList( airPlanes );
		clearFlightList( GTK_CLIST(airUi->flist) );
	}
	airObjs = freeObjList(airObjs);
	game_board_clean(airUi);
}

gboolean /* process the individual objects */
processObject(airObj *object, airSetting *settings)
{
	if( object->type->speed > 0) {
		/* the object is a mover */
		if( object->pos.heading != -1 ) {
			degsToMove( &(object->pos) );
		}
		/* something moving should be put here */
		/* move_object(object); */
	} else {
		/* the object doesn't move */
		if( object->type != &(settings->otype[ENTRY_POINT]) ) {
			move_object(object);
		}
	}
	return TRUE;
}

/* handle the plane part of a object in the game. Runs for every plane
 * every round. returns FALSE if this plane is finished for this game */
gboolean          
processPlane(airPlane *plane, airSetting *settings)
{
	airObj *exitobj= NULL;
	gboolean changed_heading = FALSE;
	gboolean planeStillInGame = TRUE;

	g_assert(plane    != NULL);
	g_assert(settings != NULL);
	
	if(plane->inTime > 0) { /* if the planes are still outside the board */
		plane->inTime -= settings->roundSpeed;
	} else if( !testVectorActive(plane->pos) ) { 
		/* first time that the plane enters the board */
		plane = planeSetup(plane);
		move_plane(plane, IMG_NORMAL, changed_heading);
//			g_print("First : %d,%d,%d\n", plane->pos.x, plane->pos.y, plane->pos.z);
	} else { /* a plane in the the field that needs to get moved */
		plane->fuel -= settings->roundSpeed;
		if( !testPlaneCourseChange(plane) ) {
				/* there was no command for the plane */
			//		g_print("There was NO command for plane %x\n", GPOINTER_TO_INT(plane));
			degsToMove( &(plane->pos) );
		} else { /* there was a command */
			/* plane's height */
			changed_heading = TRUE;
			//g_print("There was a command for plane %x\n", GPOINTER_TO_INT(plane));
			plane->pos.z       = planeChangeHeight(plane->pos.z, plane->cmd.z);
			 /* plane's turn */
			plane->pos.heading = planeTurn(plane->pos.heading, plane->cmd.heading, plane->type->turn);
			if( ( plane->cmd.z       == plane->pos.z        ) && 
			    ( plane->cmd.heading == plane->pos.heading) ) {
			/* if the command has been completed >> reset */
				plane = planeClearCmd(plane);
			}
			degsToMove( &(plane->pos) );
		}
		if( planeLeave(plane, &exitobj) ) {
			planeScore(plane, exitobj);
			remove_plane(plane);
			planeReset(plane);
			planeClearCmd(plane);
			planeStillInGame = FALSE;
		} else {
			move_plane(plane, IMG_NORMAL, changed_heading);
		}
/*		g_print("plane: x=%d, y=%d, z=%d, h=%d\n", plane->pos.x, plane->pos.y, plane->pos.z, plane->pos.heading); */
	}
	return planeStillInGame;
}

airPlane * /* set the initial values needed when the plane enters the board */
planeSetup(airPlane *plane)
{
	plane->pos.x = plane->in->pos.x;
	plane->pos.y = plane->in->pos.y;
	plane->pos.z = plane->inHeight;
	return plane;
}

airPlane * /* reset the values of the plane when the plane leaves the board */
planeReset(airPlane *plane)
{
	plane->pos.x = -1;
	plane->pos.y = -1;
	plane->pos.z = -1;
	return plane;
}


gboolean /* check if a plane left the board, return TRUE if that is the case
	  * exitobj contains a pointer to the exit object, if no object -> NULL */
planeLeave(airPlane *plane, airObj **exitobj)
{
	GList *objs = g_list_first(airObjs);
	airObj *temp = NULL;
//	exitobj = NULL;

	do { /* check all the entry points and airports */
		temp = objs->data;
		if( (plane->pos.x == temp->pos.x) && /* the positions */
		    (plane->pos.y == temp->pos.y) ) {/* of plane and
							object are the same*/
			if( temp->type == &(airSet->otype[AIRPORT]) ) {
				if(plane->pos.z == 0 ) {
					g_print("PLANE LEFT AT AIRPORT -- ");
					g_print("POS: x=%d, y=%d, z=%d, h=%d\n",
						plane->pos.x,
						plane->pos.y,
						plane->pos.z,
						plane->pos.heading);
					*exitobj = temp;	
					return TRUE;
				} else {
					exitobj = NULL;
					return FALSE;
				}
			} else if( temp->type == &(airSet->otype[ENTRY_POINT]) ) {
				g_print("PLANELEFT IN ENTRY POINT -- ");
				g_print("POS: x=%d, y=%d, z=%d, h=%d\n",
					plane->pos.x,
					plane->pos.y,
					plane->pos.z,
					plane->pos.heading);
 
				*exitobj = temp;
				return TRUE;
			} else {
				g_assert_not_reached(); /* an object that is neither a
							   plane or a entry point ? */
			}
		}
	} while( (objs = g_list_next(objs) ) != NULL);

	/* first, if the plane left the board somewhere: return*/
	if( (plane->pos.x < 0) || (plane->pos.x >= AIR_BOARD_X) ||
	    (plane->pos.y < 0) || (plane->pos.y >= AIR_BOARD_Y) ) {
		g_print("PLANE LEFT IN NORMAL PLACE -- ");
		g_print("POS: x=%d, y=%d, z=%d, h=%d\n",
			plane->pos.x,
			plane->pos.y,
			plane->pos.z,
			plane->pos.heading);
		exitobj = NULL;    /* true, the plane left the game */
		return TRUE;       /* NULL object asociated, it didn't
				      leave on an entry/exit point */
	}
	
	return FALSE;
}

void /* score the plane that leaves the board */
planeScore(airPlane *plane, airObj *exitobj)
{
	if( exitobj == plane->out ) {
		airSet->score.points += airSet->score.planeval;
		airSet->score.correct++;
		g_print("PLANE LEFT AT THE CORRECT PLACE!\n");
	} 
}

void /* gets called when planes "nearmiss" eachother */
planeNearMiss(GList *planeList) {
	airSet->score.nearmiss++;
	g_print("plane near miss warning!\n");
}

airPlane * /* reset command structure to "do nothing" values */
planeClearCmd(airPlane *plane) {
	plane->cmd.x = -1;
	plane->cmd.y = -1;
	plane->cmd.z = -1;
	plane->cmd.heading = -1;
	return plane;
}

gint /* calulate a new heading */
planeTurn(gint cur, /* current heading */
	  gint wanted, /* wanted heading */
	  gint maxturn) /* maximum amount to turn in one turn */
	
{
	gint diff;
	
	diff = cur - wanted;
	if(diff != 0) { /* if there's some turning to be done */
		if(diff >  180 ) { diff -= 360;	} 
		if(diff < -180 ) { diff += 360; }
		
		if(diff < 0) { cur += maxturn; }
		if(diff > 0) { cur -= maxturn; }
		
		/* make sure that it never gets more then 
		   360 or less then 0 */
		if(cur >= 360) { cur -= 360; }
		if(cur <    0) { cur += 360; }
	}
	return cur;

}

int /* change the height of the plane */
planeChangeHeight(int cur, /* the current height */
		  int wanted) /* the wanted height */
{
	/* going up */
	if(wanted > cur)
		cur++;
	/* going down */
	if(wanted < cur)
		cur--;
	
	return cur;
}

void /* update the horizontal and vertical movement depending
      * on the heading of the plane */
degsToMove(vector *pos)
{
//	if(pos == NULL) { die; }

	switch (pos->heading) {
	case 0:
		pos->y--;
		break;
	case 45:
		pos->x++;
		pos->y--;
		break;
	case 90:
		pos->x++;
		break;
	case 135:
		pos->x++;
		pos->y++;
		break;
	case 180:
 		pos->y++;;
		break;
	case 225:
		pos->x--;
		pos->y++;
		break;
	case 270:
		pos->x--;
		break;
	case 315:
		pos->x--;
		pos->y--;
		break;
	default:
		g_print("\nPOS->HEADING = %d\n",pos->heading);
		g_assert_not_reached();
	}
}

GList* /* check the proximity of the planes */
planeProximity(GList *planesList, 
	       gint dist)
{
	GList *badPlanesList = NULL;
	GList *curPlaneList = NULL;
	airPlane *p1 = NULL;
	airPlane *p2 = NULL;
	
	g_assert(planesList != NULL);
	g_assert(dist >= 0);

	planesList = g_list_first(planesList);
	while(planesList != NULL) {
		p1 = planesList->data;
		if( !testVectorActive(p1->pos) ) {
			planesList = g_list_next(planesList);
			continue;
		}
		curPlaneList = g_list_next(planesList); 
		/* the starting plane moves, coz we know that doesn't give conflicts. */
		/* make a list of planes that are as least as close as dist */
		while(curPlaneList != NULL) {
			p2 = curPlaneList->data;
			if( !testVectorActive(p2->pos) ) {
				curPlaneList = g_list_next(curPlaneList);
				continue;
			}
			if(p1 == p2) {
				g_assert_not_reached();
			}
//			g_print("Distance between %x and %x is %g\n",GPOINTER_TO_INT(p1),
//				GPOINTER_TO_INT(p2), objDist( p1->pos, p2->pos));
			if( objDist( p1->pos, p2->pos) <= dist) {
				badPlanesList = g_list_prepend(badPlanesList, planesList->data);
			}
			curPlaneList = g_list_next(curPlaneList);
		}
		planesList = g_list_next(planesList); 
	}
	return badPlanesList;
}

/* ChangLog
 * $Log: game.c,v $
 * Revision 1.22  2000/04/27 23:25:56  marijn
 * Clean up the game board after a game.
 *
 * Revision 1.21  2000/04/24 23:30:16  marijn
 * Added callbacks for results window. Fixed scoring bug.
 *
 * Revision 1.20  2000/04/01 16:55:47  marijn
 * Major source code reordering, added license information.
 *
 * Revision 1.19  2000/03/27 00:16:08  marijn
 * Further proximity support. Rudimentary game scoring implemented.
 *
 * Revision 1.18  2000/03/20 01:07:23  marijn
 * Various bugs fixed, plugged all known memory leaks.
 *
 * Revision 1.17  2000/03/19 19:34:14  marijn
 * Added plane distance checking, planes can now crash into eachother.
 *
 * Revision 1.16  2000/03/08 23:34:53  marijn
 * The game now has a proper ending. "End Game" added to the menu.
 *
 * Revision 1.15  2000/03/07 20:22:43  marijn
 * Planes now leave the game when they exit the board. Various minor fixes
 *
 * Revision 1.14  2000/02/21 00:17:02  marijn
 * Changed from GTK to Gnome build. Cleaned up inclusions of header-files.
 *
 * Revision 1.13  2000/01/17 01:39:22  marijn
 * Planes now keep to be the same widgets when they only move.
 * Improved color responce.
 * Added initail support for removing planes when they leave the board.
 *
 * Revision 1.12  2000/01/05 15:03:36  marijn
 * Added basic support for airport images
 *
 * Revision 1.11  1999/12/26 03:21:01  marijn
 * Planes now have xpm images for them.
 * Added comments to much of the global structures.
 *
 * Revision 1.10  1999/12/10 21:25:22  marijn
 * Added comments and cleaned gameRound() in game.c
 * Added structure for keeping track scored points.
 *
 * Revision 1.9  1999/11/18 00:12:43  marijn
 * Basic checking of where a plane leaves is now implemented.
 * GameRound is a bit more structured.
 *
 * Revision 1.8  1999/11/08 00:23:03  marijn
 * Fixed "height-loosing" bug.. planes seem to move correctly now
 *
 * Revision 1.7  1999/11/06 19:42:41  marijn
 * Buttons are disabled when no plane is selected.
 *
 * Revision 1.6  1999/09/27 21:20:14  marijn
 * Plane now turns correctly
 *
 * Revision 1.5  1999/09/26 12:30:55  marijn
 * planes turn the correct way.
 * clicking on the plane now selects it in the list
 * fixed dubble timing problem
 *
 * Revision 1.4  1999/09/17 13:45:24  marijn
 * detection that a plane leaves works
 *
 * Revision 1.3  1999/09/13 19:43:48  marijn
 * main game round code inserted..
 *
 * Revision 1.2  1999/08/27 13:29:29  marijn
 * planes are get added to the planelist
 *
 * Revision 1.1  1999/08/22 18:21:01  marijn
 * types and objects are generated as is the list of planes.
 *
 */
   
